@@ -71,6 +71,8 @@ static inline const char *ranap_domain_name(RANAP_CN_DomainIndicator_t domain)
 
 #define IUH_MSGB_SIZE	2048
 
+#define IUH_TX_QUEUE_MAX_LENGTH 1024
+
 struct hnbgw_context_map;
 
 static inline bool cnlink_is_cs(const struct hnbgw_cnlink *cnlink)
@@ -85,16 +87,19 @@ static inline bool cnlink_is_ps(const struct hnbgw_cnlink *cnlink)
 
 struct hnbgw {
 	struct {
-		const char *iuh_local_ip;
-		/*! SCTP port for Iuh listening */
-		uint16_t iuh_local_port;
 		struct osmo_plmn_id plmn;
 		uint16_t rnc_id;
-		bool hnbap_allow_tmsi;
 		/*! print hnb-id (true) or MCC-MNC-LAC-RAC-SAC (false) in logs */
 		bool log_prefix_hnb_id;
 		bool accept_all_hnb;
 		struct mgcp_client_conf *mgcp_client;
+		struct {
+			const char *local_ip;
+			/*! SCTP port for Iuh listening */
+			uint16_t local_port;
+			bool hnbap_allow_tmsi;
+			unsigned int tx_queue_max_length;
+		} iuh;
 		struct {
 			char *local_addr;
 			uint16_t local_port;
@@ -142,9 +147,7 @@ struct hnbgw {
 
 	struct {
 		struct osmo_pfcp_endpoint *ep;
-		struct osmo_pfcp_cp_peer *cp_peer;
-		/* Running counters for the PFCP conn */
-		struct osmo_stat_item_group *statg;
+		struct hnbgw_upf *upf;
 	} pfcp;
 
 	struct osmo_timer_list hnb_store_rab_durations_timer;
